"""
Generate configuration files into generated_dir.
"""

from confab.conffiles import ConfFiles
from confab.data import DataLoader
from confab.loaders import FileSystemEnvironmentLoader
from confab.validate import validate_generate

from fabric.api import task


@task
def generate(templates_dir=None, data_dir=None, generated_dir=None):
    """
    Generate configuration files.
    """
    validate_generate(templates_dir, data_dir, generated_dir)

    conffiles = ConfFiles(FileSystemEnvironmentLoader(templates_dir),
                          DataLoader(data_dir))

    conffiles.generate(generated_dir)



"""
Pull configuration files from remote host into remotes_dir.
"""

from confab.conffiles import ConfFiles
from confab.data import DataLoader
from confab.loaders import FileSystemEnvironmentLoader
from confab.validate import validate_pull

from fabric.api import task


@task
def pull(templates_dir=None, data_dir=None, remotes_dir=None):
    """
    Pull remote configuration files.
    """
    validate_pull(templates_dir, data_dir, remotes_dir)

    conffiles = ConfFiles(FileSystemEnvironmentLoader(templates_dir),
                          DataLoader(data_dir))

    conffiles.pull(remotes_dir)



"""
Jinja2 Environment loading helper functions.

Confab uses the environments list_templates() method to abstract
template location from rendering and synchronization.

Note that the default Jinja2 Loaders assume a charset (default: utf-8).
"""

from jinja2 import (Environment, FileSystemLoader, PackageLoader, BaseLoader,
                    StrictUndefined, TemplateNotFound)
from os.path import join, exists
from pkg_resources import get_provider


class FileSystemEnvironmentLoader(object):
    """Loads Jinja2 environments from directories."""

    def __init__(self, dir_name):
        self.dir_name = dir_name

    def __call__(self, component):
        """
        Load a Jinja2 Environment for a component.
        """
        component_path = join(self.dir_name, component)

        if not exists(component_path):
            return Environment(loader=EmptyLoader())

        return Environment(loader=ConfabFileSystemLoader(component_path),
                           undefined=StrictUndefined)


class PackageEnvironmentLoader(object):
    """Loads Jinja2 environments from python packages."""

    def __init__(self, package_name, templates_path='templates'):
        self.package_name = package_name
        self.templates_path = templates_path

    def __call__(self, component):
        """
        Load a Jinja2 Environment for a component.
        """
        package_path = join(self.templates_path, component)

        provider = get_provider(self.package_name)
        if not provider.resource_isdir(package_path):
            return Environment(loader=EmptyLoader())

        return Environment(loader=PackageLoader(self.package_name, package_path),
                           undefined=StrictUndefined)


class ConfabFileSystemLoader(FileSystemLoader):
    """Adds support for binary templates when loading an environment from the file system.

    Binary config files cannot be loaded as Jinja templates by default, but since confab's
    model is built around Jinja environments we need to make sure we can still represent
    them as jinja Templates.

    Since confab only renders templates from text config files (see
    :py:meth:`confab.conffiles.Conffile.generate` and :py:meth:`confab.options.should_render`)
    we can workaround this by returning a dummy template for binary config files
    with the appropriate metadata. When generating the configuration, confab,
    instead of rendering the template, will just copy the template file
    (the binary config file) verbatim to the generated folder.
    """

    def get_source(self, environment, template):

        try:
            return super(ConfabFileSystemLoader, self).get_source(environment, template)
        except UnicodeDecodeError:
            pass  # not a text file

        for searchpath in self.searchpath:
            filename = join(searchpath, template)
            if exists(filename):
                return "", filename, True

        raise TemplateNotFound(template)


class EmptyLoader(BaseLoader):
    """Jinja template loader with no templates."""

    def list_templates(self):
        return []



"""
Non-init module for doing convenient * imports from.
"""

# core
from confab.conffiles import ConfFiles

# model lading
from confab.model import load_model_from_dir, load_model_from_dict

# jinja2 environment loading
from confab.loaders import FileSystemEnvironmentLoader, PackageEnvironmentLoader

# data loading
from confab.data import DataLoader

# options
from confab.options import assume_yes, Options

# fabric tasks
from confab.diff import diff
from confab.generate import generate
from confab.pull import pull
from confab.push import push

__ignore__ = [
    assume_yes,
    ConfFiles,
    load_model_from_dir,
    load_model_from_dict,
    FileSystemEnvironmentLoader,
    PackageEnvironmentLoader,
    DataLoader,
    diff,
    generate,
    pull,
    push,
    Options
]



"""
Functions for loading configuration data.
"""
from itertools import chain
from fabric.api import puts
from jinja2 import Environment, FileSystemLoader, TemplateNotFound

from confab.files import _import, _import_string
from confab.merge import merge
from confab.options import options
from confab.output import debug


def _get_environment_module():
    """
    Return the current configuration environment.

    Placeholder.
    """
    return options.get_environmentname()


def _get_component_modules(component):
    """
    Return the different modules in the component path.
    """
    return component.split('/')


def _get_host_module():
    """
    Return the current configuration hostname.
    """
    return options.get_hostname()


def import_configuration(module_name, data_dir):
    """
    Load configuration from file as python module.

    Returns publicly names values in module's __dict__.
    """

    def as_dict(module):
        try:
            return {k: v for k, v in module.__dict__.iteritems() if not k[0:1] == '_'}
        except AttributeError:
            return None

    try:
        debug("Attempting to load {module_name}.py from {data_dir}",
              module_name=module_name,
              data_dir=data_dir)

        module = _import(module_name, data_dir)
        puts("Loaded {module_name}.py from {data_dir}".format(module_name=module_name,
                                                              data_dir=data_dir))
    except ImportError:
        debug("Attempting to load {module_name}.py_tmpl from {data_dir}",
              module_name=module_name,
              data_dir=data_dir)
        # try to load as a template
        try:
            env = Environment(loader=FileSystemLoader(data_dir))
            rendered_module = env.get_template(module_name + '.py_tmpl').render({})
        except TemplateNotFound:
            debug("Could not load {module_name} from {data_dir}",
                  module_name=module_name,
                  data_dir=data_dir)
            return None

        module = _import_string(module_name, rendered_module)
        puts("Loaded {module_name}.py_tmpl from {data_dir}".format(module_name=module_name,
                                                                   data_dir=data_dir))

    return as_dict(module)


class DataLoader(object):
    """
    Load and merge configuration data.

    Configuration data is loaded from python files by type,
    where type is defined to include defaults, per-role values,
    per-component values, per-environment values and per-host values.

    Configuration data also includes the current environment
    and host string values under a 'confab' key.
    """

    def __init__(self, data_dir):
        self.data_dir = data_dir

    def __call__(self, component):
        """
        Load the data for the given component.

        :param component: a component path, i.e. `{role}/{sub-component}/{component}`.
        """
        is_not_none = lambda x: x is not None

        module_names = filter(is_not_none,
                              chain(['default'],
                                    _get_component_modules(component),
                                    [_get_environment_module(),
                                     _get_host_module()]))

        load_module = lambda module_name: import_configuration(module_name, self.data_dir)

        module_dicts = filter(is_not_none, map(load_module, module_names))

        confab_data = dict(confab=dict(environment=options.get_environmentname(),
                                       host=options.get_hostname(),
                                       component=component))

        return merge(confab_data, *module_dicts)



"""
File options.
"""

import imp
import os
import shutil
import sys
from fabric.api import runs_once


@runs_once
def _clear_dir(dir_name):
    """
    Remove an entire directory tree.
    """
    if os.path.isdir(dir_name):
        shutil.rmtree(dir_name)


def _clear_file(file_name):
    """
    Remove an entire directory tree.
    """
    if os.path.exists(file_name):
        os.remove(file_name)


def _ensure_dir(dir_name):
    """
    Ensure that a directory exists.
    """
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)


def _import(module_name, dir_name):
    """
    Load python module from file system without reloading.

    Raises ImportError if not found.
    """

    # assign module a name that's not likely to conflict
    safe_name = 'confab.data.' + module_name

    # check if module is already loaded
    existing = sys.modules.get(safe_name)
    if existing:
        return existing

    # try to load module
    module_info = imp.find_module(module_name, [dir_name])
    module = imp.load_module(safe_name, *module_info)
    return module


def _import_string(module_name, content):
    """
    Load python module from an in-memory string without reloading.
    """

    # assign module a name that's not likely to conflict
    safe_name = 'confab.data.' + module_name

    # check if module is already loaded
    existing = sys.modules.get(safe_name)
    if existing:
        return existing

    # try to load module
    module = imp.new_module(safe_name)
    exec content in module.__dict__
    return module



"""
Configuration data structure merging functions.

We expect to obtain hierarchical configurations as native dictionaries
and want to merge the higher precedence (override) data into the
lower precedence (default) data, potentially multiple times.

Merge rules are as follows:
 - Dictionaries are merged recursively by default.
 - Lists and primitives are replaced by default.
 - Callables are called to provide custom extension.

For example, a list of hosts in the default dictionary will normally
be replaced by values defined in the override dictionary; however if
the override dictionary's list is a callable, it can be made to do
something else, such as append a new host to the default list.
"""


def _best(default_value, has_override, override_value):
    """
    Return the best value according to the merge rules.
    """
    if not has_override:
        return default_value
    elif callable(override_value):
        # custom callable
        return override_value(default_value)
    elif isinstance(default_value, dict) and isinstance(override_value, dict):
        # merge recursively
        return _merge(default_value, override_value)
    else:
        # replace with override
        return override_value


def _iterkeys(default, override):
    """
    Return an iterator over the keys of the both dicts.
    """
    return iter(set(default.keys()) | set(override.keys()))


def _entry(default, override, key):
    """
    Return a dictionary entry with the best value from the default and
    override dictionaries.
    """
    return key, _best(default.get(key), key in override, override.get(key))


def _merge(default, override):
    """Recursively merge two dictionaries.
    """
    return dict(_entry(default, override, key) for key in _iterkeys(default, override))


def merge(*args):
    """Recursively merge multiple dictionaries.
    """
    return reduce(_merge, args, {})


class Append(list):
    """
    Customized callable list that appends its values to the default.
    """
    def __init__(self, *args):
        super(Append, self).__init__(args)

    def __call__(self, default):
        return (default or []) + self


def append(*args):
    return Append(*args)


class Prepend(list):
    """
    Customized callable list that prepends its values to the default.
    """

    def __init__(self, *args):
        super(Prepend, self).__init__(args)

    def __call__(self, default):
        return self + (default or [])


def prepend(*args):
    return Prepend(*args)


class UniqueUnion(list):
    """
    Customized callable list that adds its values to the default list
    preserving unique values.
    """

    def __init__(self, *args):
        super(UniqueUnion, self).__init__(args)

    def __call__(self, default):
        return list(set(default or []) | set(*self))


def unique_union(*args):
    return UniqueUnion(*args)



"""
Resolve environment, role, and host choices into actions.

If a user specifies only an environment, confab should target all hosts
and roles in that environment. If one or more roles -- or one or more hosts --
are specified explicilty, confab should target a subset.
"""
from warnings import warn
from confab.model import get_roles_for_host, get_hosts_for_environment


def resolve_hosts_and_roles(environment, hosts=None, roles=None):
    """
    Given an environment, (possibly empty) list of hosts, and a (possibly empty)
    list of roles, return a mapping from host to roles to target.

    Raises an exception if any targeted host would have no roles after resolution.
    """
    # Validate environment
    if not environment:
        raise Exception("Environment was not specified")

    # Determine hosts for the environment
    environment_hosts = get_hosts_for_environment(environment)
    if environment_hosts is None:
        raise Exception("Environment '{}' is not recognized.".format(environment))
    elif not environment_hosts:
        warn("Environment '{}' does not have any hosts configured.".format(environment))

    # Use all environment hosts if none are specified
    hosts = hosts or environment_hosts

    # Forbid any hosts not in environment
    for host in hosts:
        if host not in environment_hosts:
            raise Exception("Host '{host}' is not a member of the environment '{env}'"
                            .format(host=host, env=environment))

    # Determine roles for hosts
    if roles:
        # If roles were specified, restrict mapping to those roles
        valid_roles = lambda role: role in roles
        hosts_to_roles = {}
        for host in hosts:
            host_roles = filter(valid_roles, get_roles_for_host(host))
            if host_roles:
                hosts_to_roles[host] = host_roles
    else:
        # Otherwise, use all roles
        hosts_to_roles = dict([(host, get_roles_for_host(host)) for host in hosts])

    # Validate that all hosts have at least one role
    for host, host_roles in hosts_to_roles.iteritems():
        if not host_roles:
            if roles:
                raise Exception("Host '{}' does not have any of the specified roles".format(host))
            else:
                raise Exception("Host '{}' does not have any configured roles".format(host))

    return hosts_to_roles



"""
Functions for interacting with the defined hosts, environments, and roles.
"""
from fabric.api import env
from confab.files import _import
from os.path import join


def _keys():
    """
    Get the model's environment keys.
    """
    return ["environmentdefs", "roledefs", "componentdefs"]


def load_model_from_dir(dir_name, module_name='settings'):
    """
    Load model data (environments, roles, hosts) from settings module.
    """

    settings = _import(module_name, dir_name)
    for key in _keys():
        env[key] = getattr(settings, key, {})


def load_model_from_dict(settings):
    """
    Load model data (environments, roles, hosts) from settings dictionary.
    """

    for key in _keys():
        env[key] = settings.get(key, {})


def get_roles_for_host(host):
    """
    Get all roles that a host belongs to.

    Delegates to Fabric's env roledefs.
    """
    return [role for (role, hosts) in env.roledefs.iteritems() if host in hosts]


def get_hosts_for_environment(environment):
    """
    Get all hosts for an environment.

    Assumes an environmentsdef structure in Fabric's env.
    """
    try:
        return env.environmentdefs[environment]
    except AttributeError:
        raise Exception("No environments are defined")
    except KeyError:
        raise Exception("Environment '{}' is not defined".format(environment))


def get_components_for_role(role):
    """
    Get all component paths for the given role.
    """
    if role not in env.roledefs:
        raise Exception("Role '{}' is not defined".format(role))

    if role not in env.componentdefs:
        return []

    return _expand_components(role, '', {})


def _expand_components(component, path, seen):
    """
    Recursively expand component paths rooted at the given component
    based on componentdefs.

    Raises an exception if a cycle is discovered in component definitions
    or if a component leaf exists in multiple paths.

    :param component: A component name.
    :param path: The component path being built.
    :param seen: Dictionary of visited components and their paths.
    """
    component_path = join(path, component)

    if component in seen:
        raise Exception("Detected cycle or multiple paths with role/component '{}'"
                        " ('{}' and '{}')".format(component,
                                                  seen[component],
                                                  component_path))
    seen[component] = component_path

    if component not in env.componentdefs:
        return [component_path]

    components = []
    for c in env.componentdefs.get(component):
        components += _expand_components(c, component_path, seen)

    return components



"""
Push generated configuration files to remote host.
"""

from confab.conffiles import ConfFiles
from confab.data import DataLoader
from confab.loaders import FileSystemEnvironmentLoader
from confab.validate import validate_all

from fabric.api import task


@task
def push(templates_dir=None,
         data_dir=None,
         generated_dir=None,
         remotes_dir=None):
    """
    Push configuration files.
    """
    validate_all(templates_dir, data_dir, generated_dir, remotes_dir)

    conffiles = ConfFiles(FileSystemEnvironmentLoader(templates_dir),
                          DataLoader(data_dir))

    conffiles.push(generated_dir, remotes_dir)



"""
Functions for validating user input to tasks.
"""

from fabric.api import abort, env
import os


def validate_templates_dir(templates_dir):
    """
    Template directory must be defined and exist.
    """
    if not templates_dir:
        abort('Please provide a valid templates directory')
    elif not os.path.isdir(templates_dir):
        abort('{templates_dir} is not a valid directory'.format(templates_dir=templates_dir))


def validate_data_dir(data_dir):
    """
    Data directory must be defined and exist.
    """
    if not data_dir:
        abort('Please provide a valid data_dir')
    elif not os.path.isdir(data_dir):
        abort('{data_dir} is not a valid directory'.format(data_dir=data_dir))


def validate_generated_dir(generated_dir):
    """
    Generated directory must be defined and not be a regular file.
    """
    if not generated_dir:
        abort('Please provide a valid generated_dir')
    elif os.path.exists(generated_dir) and not os.path.isdir(generated_dir):
        abort('{generated_dir} is not a valid directory'.format(generated_dir=generated_dir))


def validate_remotes_dir(remotes_dir):
    """
    Remotes directory must be defined and not be a regular file.
    """
    if not remotes_dir:
        abort('Please provide a valid remotes_dir')
    elif os.path.exists(remotes_dir) and not os.path.isdir(remotes_dir):
        abort('{remotes_dir} is not a valid directory'.format(remotes_dir=remotes_dir))


def validate_host():
    """
    Fabric host_string must be defined.
    """
    if not env.host_string:
        abort('Please specify a host or a role')


def validate_all(templates_dir, data_dir, generated_dir, remotes_dir):
    """
    Validate templates_dir, data_dir, generated_dir, remotes_dir, and host.
    """
    validate_templates_dir(templates_dir)
    validate_data_dir(data_dir)
    validate_generated_dir(generated_dir)
    validate_remotes_dir(remotes_dir)
    validate_host()


def validate_generate(templates_dir, data_dir, generated_dir):
    """
    Validate templates_dir, data_dir, generated_dir, and host.
    """
    validate_templates_dir(templates_dir)
    validate_data_dir(data_dir)
    validate_generated_dir(generated_dir)
    validate_host()


def validate_pull(templates_dir, data_dir, remotes_dir):
    """
    Validate templates_dir, data_dir, remotes_dir, and host.
    """
    validate_templates_dir(templates_dir)
    validate_data_dir(data_dir)
    validate_remotes_dir(remotes_dir)
    validate_host()



"""
Auto-generate Fabric tasks for roles and environments.

The 'autogenerate_tasks' function creates fabric tasks to
set the current role or environment and are intended to be
used along side the other standard Confab tasks (e.g. pull)
to customize configuration data.

These tasks are somewhat experimental.
"""
from fabric.api import env
from fabric.state import commands


def _add_task(name, task, doc):
    """
    Register a Fabric task.
    """
    task_wrapper = lambda: task()
    setattr(task_wrapper, '__doc__', doc)
    commands[name] = task_wrapper


def autogenerate_tasks():
    """
    Autogenerate role_ and env_ tasks for all defined roles and
    environments in the Fabric environment.

    Normally the roledefs and environment defs will be configured
    using 'load_model_from_dir' or similar.
    """
    for role in env.roledefs.iterkeys():
        _add_task('role_' + role,
                  lambda: setattr(env, 'role', role),
                  "Set role to '{role}'".format(role=role))

    try:
        for environment in env.environmentdefs.iterkeys():
            _add_task('env_' + environment,
                      lambda: setattr(env, 'environment', environment),
                      "Set environment to '{environment}'".format(environment=environment))
    except AttributeError:
        # environment defs not configured
        pass



"""
Options for managing Confab.
"""

from fabric.api import env, task
from fabric.utils import _AttributeDict

from difflib import unified_diff
from magic import Magic
from re import match


def _should_render(mime_type):
    """
    Return whether a template file of with a particular mime type
    should be rendered.

    Some files may need to be excluded from template rendering;
    such files will be copied verbatim.
    """
    return next((True for pattern in ['text/', 'application/xml'] if match(pattern, mime_type)),
                False)


def _is_empty(mime_type):
    """
    Return whether a template file is an empty file.
    """
    return mime_type == 'inode/x-empty'


def _is_not_temporary(file_name):
    """
    Return whether a file name does not represent a temporary file.

    When listing configuration files, we usually want temporary
    files to be ignored.
    """
    return not file_name.endswith('~')


def _get_mime_type(file_name):
    """
    Return the mime type of a file.

    The mime_type will be used to determine if a configuration file is text.
    """
    return Magic(mime=True).from_file(file_name)


def _get_hostname():
    """
    Return the current target hostname.
    """
    return env.host_string


def _get_rolename():
    """
    Return the current target role.

    Assume the role value is being saved in Fabric's env;
    if not return None.
    """
    try:
        return env.role
    except AttributeError:
        return None


def _get_environmentname():
    """
    Return the current target environment.

    Assume the environment value is being saved in Fabric's env;
    if not return None.
    """
    try:
        return env.environment
    except AttributeError:
        return None


def _diff(a, b, fromfile=None, tofile=None):
    """
    Return a diff using '---', '+++', and '@@' control lines.

    By default, uses unified_diff.
    """
    return unified_diff(a, b, fromfile=fromfile, tofile=tofile)

# Options that control how confab runs.
#
# These are in opposition to options likely to changed
# between different runs of confab, such as directories,
# environments, roles, hosts, etc.
options = _AttributeDict({
    # Should yes be assumed for interactive prompts?
    'assume_yes': False,

    # How to compute a file's mime_type?
    'get_mime_type': _get_mime_type,

    # How to determine if a template should be rendered?
    'should_render': _should_render,

    # How to determine if a template is an empty file?
    'is_empty': _is_empty,

    # How to determine the current host name?
    'get_hostname': _get_hostname,

    # How to determine the current role name?
    'get_rolename': _get_rolename,

    # How to determine the current environment name?
    'get_environmentname': _get_environmentname,

    # How do filter available templates within the jinja environment?
    'filter_func': _is_not_temporary,

    # How to determine diffs?
    'diff': _diff
})


class Options(object):
    """
    Context manager to temporarily set options.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.previous = {}

    def __enter__(self):
        for (k, v) in self.kwargs.iteritems():
            self.previous[k] = options.get(k)
            options[k] = v
        return self

    def __exit__(self, exc_type, value, traceback):
        for k in self.kwargs.keys():
            options[k] = self.previous[k]


@task
def assume_yes():
    """
    Set the option to assume_yes in other tasks.
    """
    options.assume_yes = True



#!/usr/bin/env python

from setuptools import setup, find_packages

# Workaround for running "setup.py test"
# See: http://bugs.python.org/issue15881
try:
    import multiprocessing # flake8: NOQA
except ImportError:
    pass

__version__ = '1.1'

# Jenkins will replace __build__ with a unique value.
__build__ = ''

setup(name='confab',
      version=__version__ + __build__,
      description='Configuration management with Fabric and Jinja2.',
      author='Location Labs',
      author_email='info@locationlabs.com',
      url='http://github.com/locationlabs/confab',
      license='Apache2',
      packages=find_packages(exclude=['*.tests']),
      setup_requires=[
          'nose>=1.0'
      ],
      install_requires=[
          'Fabric>=1.4',
          'Jinja2>=2.4',
          'python-magic'
      ],
      test_requires=[
          'mock==1.0.1'
      ],
      test_suite='confab.tests',
      entry_points={
          'console_scripts': [
              'confab = confab.main:main',
          ]
      },
      )



"""
Configuration file template object model.
"""
from warnings import warn
from fabric.api import get, put, settings, sudo
from fabric.colors import blue, red, green, magenta
from fabric.contrib.files import exists
from fabric.contrib.console import confirm

from confab.files import _clear_dir, _clear_file, _ensure_dir
from confab.options import options
from confab.model import get_components_for_role, get_roles_for_host
from confab.output import status

import os
import shutil


class ConfFileDiff(object):
    """
    Encapsulation of the differences between the (locally copied) remote and
    generated versions of a configuration file.
    """

    def __init__(self, remote_file_name, generated_file_name, conffile_name):
        """
        Compute whether the conffile with the given name has changed given
        a remote and generate file copy.
        """
        self.missing_generated = False
        self.missing_remote = False
        self.conffile_name = conffile_name
        self.diff_lines = []

        if not os.path.exists(generated_file_name):
            # Unexpected
            self.missing_generated = True

        if not os.path.exists(remote_file_name):
            self.missing_remote = True

        if not self.missing_generated and not self.missing_remote:
            diff_iter = options.diff(open(remote_file_name).readlines(),
                                     open(generated_file_name).readlines(),
                                     fromfile='{file_name} (remote)'.format(file_name=conffile_name),
                                     tofile='{file_name} (generated)'.format(file_name=conffile_name))

            # unified_diff returns a generator
            self.diff_lines = list(diff_iter)

    def show(self):
        """
        Print the diff using pretty colors.

        If confab is used on binary files, diffs are likely to render poorly.
        """
        if self.missing_generated:
            # Unexpected
            if not self.missing_remote:
                print(red('Only in remote: {file_name}'.format(file_name=self.conffile_name)))
        elif self.missing_remote:
            print(blue(('Only in generated: {file_name}'.format(file_name=self.conffile_name))))
        else:
            for diff_line in self.diff_lines:
                color = red if diff_line.startswith('-') else blue if diff_line.startswith('+') else green
                print(color(diff_line.strip()))

    def __nonzero__(self):
        """
        Evaluate to true if there is a diff.
        """
        if self.missing_generated:
            return not self.missing_remote
        elif self.missing_remote:
            return True
        else:
            return len(self.diff_lines)


class ConfFile(object):
    """
    Encapsulation of a configuration file template.
    """

    def __init__(self, template, data):
        self.template = template
        self.data = data
        self.mime_type = options.get_mime_type(template.filename)
        self.name = template.environment.from_string(template.name).render(**self.data)
        self.remote = os.sep + self.name

    def _write_verbatim(self, generated_file_name):
        """
        Write the configuration file without templating.
        """
        shutil.copy2(self.template.filename, generated_file_name)

    def _write_template(self, generated_file_name):
        """
        Write the configuration file as a template.
        """
        with open(generated_file_name, 'w') as generated_file:
            rendered = self.template.render(**self.data).encode('utf-8')
            generated_file.write(rendered)
            generated_file.write(u'\n')
            shutil.copystat(self.template.filename, generated_file_name)

    def diff(self, generated_dir, remotes_dir, output=False):
        """
        Compute the diff between the generated and remote files.

        If output is enabled, show the diffs nicely.
        """
        generated_file_name = os.sep.join([generated_dir, self.name])
        remote_file_name = os.sep.join([remotes_dir, self .name])

        status('Computing diff for {file_name}', file_name=self.remote)

        return ConfFileDiff(remote_file_name, generated_file_name, self.remote)

    def should_render(self):
        return options.should_render(self.mime_type)

    def is_empty(self):
        return options.is_empty(self.mime_type)

    def generate(self, generated_dir):
        """
        Write the configuration file to the dest_dir.
        """
        generated_file_name = os.sep.join([generated_dir, self.name])

        status('Generating {file_name}', file_name=self.remote)

        # ensure that destination directory exists
        _ensure_dir(os.path.dirname(generated_file_name))

        if self.should_render():
            self._write_template(generated_file_name)
        else:
            self._write_verbatim(generated_file_name)

    def pull(self, remotes_dir):
        """
        Pull remote configuration file to local file.
        """
        local_file_name = os.sep.join([remotes_dir, self.name])

        status('Pulling {file_name} from {host}',
               file_name=self.remote,
               host=options.get_hostname())

        _ensure_dir(os.path.dirname(local_file_name))
        _clear_file(local_file_name)

        with settings(use_ssh_config=True):
            if exists(self.remote, use_sudo=True):
                get(self.remote, local_file_name)
            else:
                status('Not found: {file_name}',
                       file_name=self.remote)

    def push(self, generated_dir):
        """
        Push the generated configuration file to the remote host.
        """
        generated_file_name = os.sep.join([generated_dir, self.name])
        remote_dir = os.path.dirname(self.remote)

        status('Pushing {file_name} to {host}',
               file_name=self.remote,
               host=options.get_hostname())

        with settings(use_ssh_config=True):
            sudo('mkdir -p {dir_name}'.format(dir_name=remote_dir))

            put(generated_file_name,
                self.remote,
                use_sudo=True,
                mirror_local_mode=True)


class ConfFiles(object):
    """
    Encapsulation of a set of configuration files.
    """

    def __init__(self, environment_loader, data_loader):
        """
        On init, load a list of configuration files using the provided Jinja2
        environment loader and data loader.

        The environment loader must return a Jinja2 environment that uses a
        loader that supports list_templates().
        """
        def load_templates(component):

            data = data_loader(component)
            environment = environment_loader(os.path.basename(component))

            conffiles = []
            for conffile in map(lambda template_name:
                                ConfFile(environment.get_template(template_name), data),
                                environment.list_templates(filter_func=options.filter_func)):

                other_component = conffile_names.get(conffile.name)
                if other_component:
                    raise Exception("Found two configuration templates with the same file path."
                                    " File: {}, Components: {}"
                                    .format(conffile.remote,
                                            ",".join(component, other_component)))

                conffile_names[conffile.name] = component

                # store the conffiles for the current role
                if role == options.get_rolename():
                    conffiles.append(conffile)

            return conffiles

        self.conffiles = []

        # Make sure we can generate all templates for the current host.
        # This will go over all roles for the host.

        # A mapping between a conffile name
        # and the component it is generated for.
        conffile_names = {}

        for role in get_roles_for_host(options.get_hostname()):
            self.conffiles.extend(load_templates(role))
            for component in get_components_for_role(role):
                self.conffiles.extend(load_templates(component))

        if not self.conffiles:
            warn("No conffiles found for '{role}' on '{host}' in environment '{environment}'"
                 .format(role=options.get_rolename(),
                         host=options.get_hostname(),
                         environment=options.get_environmentname()))

    def generate(self, generated_dir):
        """
        Write all configuration files to generated_dir.
        """
        host_generated_dir = os.sep.join([generated_dir, options.get_hostname()])

        _clear_dir(host_generated_dir)
        _ensure_dir(host_generated_dir)

        for conffile in self.conffiles:
            conffile.generate(host_generated_dir)

    def pull(self, remotes_dir):
        """
        Pull remote versions of files into remotes_dir.
        """
        host_remotes_dir = os.sep.join([remotes_dir, options.get_hostname()])

        for conffile in self.conffiles:
            conffile.pull(host_remotes_dir)

    def diff(self, generated_dir, remotes_dir):
        """
        Show diffs for all configuration files.
        """
        host_generated_dir = os.sep.join([generated_dir, options.get_hostname()])
        host_remotes_dir = os.sep.join([remotes_dir, options.get_hostname()])

        for conffile in self.conffiles:
            conffile.pull(host_remotes_dir)

        for conffile in self.conffiles:
            conffile.generate(host_generated_dir)

        for conffile in self.conffiles:
            conffile.diff(host_generated_dir, host_remotes_dir).show()

    def push(self, generated_dir, remotes_dir):
        """
        Push configuration files that have changes, given user confirmation.
        """
        host_generated_dir = os.sep.join([generated_dir, options.get_hostname()])
        host_remotes_dir = os.sep.join([remotes_dir, options.get_hostname()])

        for conffile in self.conffiles:
            conffile.pull(host_remotes_dir)

        for conffile in self.conffiles:
            conffile.generate(host_generated_dir)

        has_diff = lambda conffile: conffile.diff(host_generated_dir, host_remotes_dir, True)
        with_diffs = filter(has_diff, self.conffiles)

        if not with_diffs:
            print(magenta('No configuration files to push for {host}'
                          .format(host=options.get_hostname())))
            return

        print(magenta('The following configuration files have changed for {host}:'
                      .format(host=options.get_hostname())))
        print
        for conffile in with_diffs:
            print(magenta('\t' + conffile.remote))

        if options.assume_yes or confirm('Push configuration files to {host}?'
                                         .format(host=options.get_hostname()),
                                         default=False):
            for conffile in with_diffs:
                conffile.push(host_generated_dir)



"""
Main function declaration for confab console_script.

Confab may be used from within a fabfile or as a library. The main
function here is provided as a simple default way to invoke confab's
tasks:

 -  A single directory root is assumed, with templates, data, generated
    and remotes directories defined as subdirectories.

 -  A host list must be provided on the command line.

For more complex invocation, a custom fabfile may be more appropriate.
"""
import getpass
import os
import sys
from optparse import OptionParser
from fabric.api import settings
from fabric.network import disconnect_all

from confab.api import pull, push, diff, generate
from confab.resolve import resolve_hosts_and_roles
from confab.model import load_model_from_dir
from confab.options import Options
from confab.output import configure_output, status


_tasks = {"diff":     (diff,     True,  True),
          "generate": (generate, True,  False),
          "pull":     (pull,     False, True),
          "push":     (push,     True,  True)}


def parse_options():
    """
    Parse command line options.

    Directory and host are required, though directory defaults to the current
    working directory.
    """

    usage = "confab [options] {tasks}".format(tasks="|".join(_tasks.keys()))
    parser = OptionParser(usage=usage)

    parser.add_option("-d", "--directory", dest="directory",
                      default=os.getcwd(),
                      help="directory from which to load configuration [default: %default]")

    parser.add_option("-e", "--environment", dest="environment",
                      default="local",
                      help="environment to operate on [default: %default]")

    parser.add_option("-H", "--hosts", dest="hosts",
                      default="",
                      help="comma-separated list of hosts to operate on")

    parser.add_option("-q", "--quiet", dest="quiet",
                      action="store_true",
                      default=False,
                      help="minimize output verbosity")

    parser.add_option("-R", "--roles", dest="roles",
                      default="",
                      help="comma-separated list of roles to operate on")

    parser.add_option("-u", "--user", dest="user",
                      default=getpass.getuser(),
                      help="username to use when connecting to remote hosts")

    parser.add_option("-v", "--verbose", dest="verbosity",
                      action="count",
                      default=0,
                      help="control confab verbosity; by default, confab suppresses"
                      "most output. Additional -v flags increase verbosity.")

    parser.add_option("-y", "--yes", dest="assume_yes",
                      action="store_true",
                      default=False,
                      help="automatically answer yes to prompts")

    opts, args = parser.parse_args()
    return parser, opts, args


def main():
    """
    Main command line entry point.
    """
    try:
        # Parse and validate arguments
        parser, options, arguments = parse_options()

        configure_output(options.verbosity,
                         options.quiet)

        try:
            load_model_from_dir(options.directory)
        except ImportError as e:
            parser.error("Unable to load {settings}: {error}"
                         .format(settings=os.path.join(options.directory, "settings.py"),
                                 error=e))

        # Normalize and resolve hosts to roles mapping
        try:
            hosts_to_roles = resolve_hosts_and_roles(options.environment,
                                                     options.hosts.split(",") if options.hosts else [],
                                                     options.roles.split(",") if options.roles else [])
        except Exception as e:
            parser.error(e)

        # Determine task
        try:
            task_name = arguments[0]
            (task, needs_templates, needs_remotes) = _tasks[task_name]
        except IndexError:
            parser.error("Please specify a task")
        except KeyError:
            parser.error("Specified task must be one of: {tasks}"
                         .format(tasks=", ".join(_tasks.keys())))

        # Construct task arguments
        kwargs = {"data_dir": os.path.join(options.directory, "data")}

        if needs_templates:
            kwargs["generated_dir"] = os.path.join(options.directory, "generated")
        if needs_remotes:
            kwargs["remotes_dir"] = os.path.join(options.directory, "remotes")

        kwargs["templates_dir"] = os.path.join(options.directory, "templates")

        # Invoke task once per host/role
        for host, roles in hosts_to_roles.iteritems():
            for role in roles:
                with settings(environment=options.environment,
                              host_string=host,
                              role=role,
                              user=options.user):
                    status("Running {task} for '{env}' and '{role}'",
                           task=task_name,
                           env=options.environment,
                           role=role)
                    with Options(assume_yes=options.assume_yes):
                        task(**kwargs)

    except SystemExit:
        raise
    except KeyboardInterrupt:
        sys.stderr.write("\nInterrupted\n")
        sys.exit(1)
    except:
        sys.excepthook(*sys.exc_info())
        sys.exit(1)
    finally:
        disconnect_all()
    sys.exit(0)



"""
Output control utilities.
"""
import warnings
from fabric.api import env
from fabric.state import output
from fabric.utils import warn


def status(message, **kwargs):
    """
    Generate fabric-style output if and only if status output
    has been selected.
    """
    if not output.status:
        return
    print "[{hostname}] {message}".format(hostname=env.host_string,
                                          message=message.format(**kwargs))


def debug(message, **kwargs):
    """
    Generate fabric-style output if and only if debug output
    has been selected.
    """
    if not output.debug:
        return
    print "[{hostname}] {message}".format(hostname=env.host_string,
                                          message=message.format(**kwargs))


def warn_via_fabric(message, category, filename, lineno=None, line=None):
    """
    Adapt Python warnings to Fabric's warning output manager.
    """
    warn(message.message)


def configure_output(verbosity=0, quiet=False):
    """
    Configure verbosity level through Fabric's output managers.
    """
    output_levels = {
        "status": 0,
        "aborts": 0,
        "user": 1,
        "warnings": 1,
        "running": 1,
        "stdout": 2,
        "stderr": 2,
        "debug": 2
    }

    verbosity = verbosity if not quiet else -1

    # configure output manager levels via verbosity
    for manager, level in output_levels.iteritems():
        output[manager] = level <= verbosity

    # Hook up Python warnings to warning output manager
    warnings.showwarning = warn_via_fabric
    warnings.simplefilter("always", UserWarning)
